from .cli import command, arg

@command('export',
    arg('--format', choices=('zip', 'dir'), default="dir"),
    arg('index_name', help="name of index to export"),
    )
def cmd_export(config):
    """Export indices into your working directory
    """
    from .connection import elasticsearch
    es = elasticsearch(config)
    from elasticsearch.helpers import scan

    index_name = config.get('export.index_name')
    result = es.indices.get(index_name)

    for idx_name, idx_data in result.items():

        if config.get('export.format') == 'dir':
            os.makedirs(idx_name)
            with open(join(idx_name, 'mappings.json'), 'w') as f:
                json.dump(result[idx_name]['mappings'], f)

            log.info("exporting %s to %s/", idx_name, idx_name)

            with open(join(idx_name, 'data.json'), 'w') as f:
                for data in scan(es, query={'query': {'match_all': {}}}, index=index_name):
                    json.dump(data, f)
                    f.write("\n")

        elif config.get('export.format') == 'zip':
            pass


@command('import', arg('index_name'))
def cmd_import(config):
    index_name = config.get('index_name')

    with open(join(index_name, 'mappings.json'), 'r') as f:
        mappings = json.load(f)

    es.indices.create(index_name, body={'mappings': mappings})
    with open(join(index_name, 'data.json'), 'r') as f:
        bulk(es, [ json.loads(line) for line in f ], refresh=True)



from elasticsearch import Elasticsearch
from .util import get_netrc_login_data
import logging
log = logging.getLogger('elastico.connection')

def elasticsearch(config={}):
    # elasticsearch configuration
    elasticsearch = config.get('elasticsearch', {})

    try:
        (user, password) = get_netrc_login_data(config, 'elasticsearch.netrc')
        elasticsearch['http_auth'] = (user, password)
        del elasticsearch['netrc']
    except LookupError:
        try:
            (user, password) = get_netrc_login_data(config, 'netrc')
            elasticsearch['http_auth'] = (user, password)
        except LookupError:
            pass

    log.debug("elasticsearch: %s", elasticsearch)

    es = Elasticsearch(**elasticsearch)
    return es



import time
import logging
import sys
from datetime import datetime, timedelta
from .notifier import Notifier
from .config import Config
from .util import to_dt, dt_isoformat, get_alerts

log = logging.getLogger('elastico.server')

class Server:
    '''This is a simple class for a server.
    '''

    def __init__(self, config, prefix=None, run=None):
        self.config = config
        self.prefix = prefix
        self.func   = run

        before_5s = to_dt(datetime.utcnow() - timedelta(seconds=5))
        if before_5s < to_dt(self.config.get('at')) < to_dt(datetime.utcnow()):
            self.run_now = True
        else:
            self.run_now = False

    def get_value(self, name, default=None):
        if self.prefix:
            return self.config.get('%s.%s' % (prefix, name), default)
        else:
            return self.config.get(name, default)

    def run(self, count=None, sleep_seconds=None):
        counter = 0
        error_count = 0
        while True:
            start = datetime.utcnow()

            if self.run_now:
                _at = dt_isoformat(to_dt(start), timespec='seconds')
            else:
                _at = dt_isoformat(
                    to_dt(self.config.get('at')) +
                    timedelta(seconds=int(sleep_seconds))
                    )

            self.config.refresh(at=_at)
            self.config.logging_setup()

            if count is None:
                count = int(self.get_value('serve.count', -1))
            if sleep_seconds is None:
                sleep_seconds = float(self.get_value('serve.sleep_seconds', 60))

            log.info("run -- counter=%r, count=%r, sleep_seconds=%r, at=%r",
                counter, count, sleep_seconds, _at)

            if count > 0:
                if counter >= count:
                    break

            try:
                self.func()
                error_count = 0
            except Exception as e:
                import traceback
                error_count += 1

                log.error("fatal error running server function -- "
                    "message=%r error_count=%r, args=%r", e, error_count, e.args[1:], exc_info=1)

                errors_max = self.get_value('serve.errors_max')
                check_errors_max = errors_max is not None and errors_max > -1

                notifier = Notifier(self.config, prefixes=[self.prefix])
                alerts = get_alerts(self.get_value('serve.alerts', []), context=self.config)
                for alert in alerts:
                    if error_count >= alert.get('error_count', 1):
                        notify = alert.get('notify', [])
                        subject = '[elastico] %s -- exception in server function' % alert.get('type', 'error')

                        if check_errors_max and error_count > errors_max:
                            subject = '[elastico] too many errors, giving up' % error_count

                        notifier.notify(notify=notify, data=Config({
                            'message': {
                                'subject': subject,
                                'text': "error_count=%s\nargs=%r\n\n" % (error_count, args) +
                                    traceback.format_exc()
                            }
                        }))

                if check_errors_max and error_count > errors_max:
                    sys.exit(1)

            duration = datetime.utcnow() - start
            log.info("run -- took %s", duration)

            sleep_time = (timedelta(seconds=sleep_seconds) - duration)
            sleep_time = sleep_time.total_seconds()
            if sleep_time < 0:
                log.warning("run took %ss longer than sleep_time", sleep_time*-1)
                sleep_time = 0

            time.sleep(sleep_time)
            counter += 1



import pyaml, yaml

from .util import PY3

from textwrap import wrap

from pyaml import UnsafePrettyYAMLDumper

def represent_stringish(dumper, data):
    data = unicode(data) # read the comment above

    style = dumper.pyaml_string_val_style
    if not style:
        style = 'plain'

        multiline = wrap(data, width=dumper.best_width)
        if "\n" not in data.strip() and len(multiline) > len(data.splitlines()):
            style = ">"

        elif '\n' not in data and ('@' in data or '{' in data or '}' in data):
            style = "'"

        elif '\n' in data or not data or data == '-' or data[0] in '!&*[?':
            style = "|"

    return yaml.representer.ScalarNode('tag:yaml.org,2002:str', data, style=style)

if PY3:
    unicode = str
for str_type in {bytes, unicode}:
    UnsafePrettyYAMLDumper.add_representer(
        str_type, represent_stringish )




import requests
import os
import json
import logging
import pyaml

from argdeco import command, main, arg, opt, config_factory
from os.path import exists, join
from zipfile import ZipFile, ZipInfo

from .connection import elasticsearch
from .search import build_search_body
from .config_factory import ConfigFactory
from .config import Config

# initialize logger
logger = logging.getLogger('elastico.cli')

# use our compiler factory for generating config object
main.configure(compiler_factory=config_factory(Config,
    prefix = 'arguments',
    config_file=arg( '--config-file', '-C',
        help="File to read configuration from"),
    ))


# add global arguments
main.add_arguments(
    arg('--host', '-H', help="Elasticsearch host. (CFG: elasticsearch.hosts)", config="elasticsearch.hosts"),
    arg('--netrc', help="get netrc entry <machine>. (CFG: netrc.machine)", config="netrc.machine"),
    arg('--netrc-file', help="set netrc file to read. (CFG: netrc.file)", config="netrc.file"),
)


class MyZipFile(ZipFile):
    '''This class overrides :py:meth:~zipfile.ZipFile:'s ``_extract_member``
    method to set executable flags, if set in ZIP file'''

    def _extract_member(self, member, targetpath, pwd):
        if not isinstance(member, ZipInfo):
            member = self.getinfo(member)

        if targetpath is None:
            targetpath = os.getcwd()

        ret_val = ZipFile._extract_member(self, member, targetpath, pwd)
        attr = member.external_attr >> 16
        os.chmod(ret_val, attr)
        return ret_val


@command('install')
def install(config):
    '''install elasticsearch from zip right here, mostly used for testing'''

    package = "elasticsearch"
    version = "6.3.2"
    filename = "{}-{}.zip".format(package, version)

    if not exists(filename):
        url = "https://artifacts.elastic.co/downloads/elasticsearch/"+filename

        logger.info("Downloading %s", url)

        r = requests.get(url)
        with open(filename, 'wb') as output:
            for chunk in r.iter_content(chunk_size=512*1024):
                output.write(chunk)

    logger.info("Extracting %s", filename)
    import zipfile
    zip_ref = MyZipFile(filename, 'r')
    zip_ref.extractall(".")
    zip_ref.close()

@command('run')
def install(config):
    """run elastic search in current directory"""

    package = "elasticsearch"
    version = "6.3.2"

    executable = "{}-{}/bin/elasticsearch".format(package, version)

    #if not exists(executable)

    os.execl(executable, executable)

@command('search',
    arg('--format', '-F',
        help="format string, which is applied to each match",
        default=None),
    arg('query', help="may be a query, a filename or '-' to read from stdin"),
)
def search(config): #host, format, query):
    pyaml.p(config)
    return 0

    format = config.get('search.format')
    query  = config.get('search.query')

    content = None
    if query == '-':
        content = sys.stdin.read()
    elif exists(query):
        with open(query, 'r') as f:
            content = f.read()
    else:
        config = {}
        config['query'] = query
        config['format'] = format

    if content is not None:
        if content.startswith("---"):
            import frontmatter
            config, content = frontmatter.loads(content)
            config['template'] = content
        else:
            config = yaml.loads(io.StringIO(content))

    body = build_search_body(config, 'search')
    es = elasticsearch(config)
    results = es.search(index=config.get('index', '*'), body=body)

    if config['format']:
        for hit in results['hits']['hits']:
            print(config['format'].format(hit))

    elif config['template']:
        import chevron
        args = {}
        for k in ('template', 'partials_path', 'partials_ext',
           'partials_dict', 'padding', 'def_ldel', 'def_rdel'):
            if k in config:
                args[k] = config[k]

        data = config.get('data', {}).copy()
        data.update(results)
        args['data'] = data
        print(chevron.render(**args))

    else:
        pyaml.p(results)


# alert commands

from .alert import Alerter

alert_command = command.add_subcommands('alert',)
arg_config = arg('config', help="configuration file or '-' to read from stdin", default=None)

def read_config(config):
    if not config:
        return {}

    if config == '-':
        config = yaml.load(sys.stdin)
    elif exists(config):
        with open(config, 'r') as f:
            config = yaml.load(f)

    path = config.get('rules_path')
    if 'rules' not in config:
        config['rules'] = []

    if path:
        path = path.format(**config)
        read_config_dir(path.format(**config), config, 'rules')

    config['arguments'] = {}
    return config


alert_command("expand_rules",
    arg_config,
    description = Alerter.expand_rules.__doc__
    )
def cmd_alert_expand_rules(config):
    config_factory = ConfigFactory(config)

#    config = read_config(config)

    for r in Alerter.expand_rules(config_factory):
        print("---")
        pyaml.p(r)


alert_command("run",
    arg('--dry-run', help="do a dry run without writing to index"),
    arg_config,
    description = Alerter.run.__doc__
    )
def cmd_alert_run(config):
    cfg = read_config(config['config'])
    cfg['arguments'] = config
    Alerter.run(cfg)

digest_command = command.add_subcommands('digest',)

'''idea of digest commands is to analyse and do digestions.

Here an example for aggregating filesystem metricset of metric beat:

Configuration:
```yaml
# for each index matching index pattern
retentions:
- index-pattern: metricbeat-*

  target: history-metricbeat-%Y-%m

  digest:
  - name: metricset-filesystem

    fields:
        metricset.module: "system"
        metricset.name: "filesystem"

    buckets:
        terms:
            tags:        tags
            host_name:   host.name
            mount_point: system.filesystem.mount_point

    aggregate:
        terms:
            metricset.name:   metricset.name
            metricset.module: metricset.module
        stats:
            available:  system.filesystem.available
            free:       system.filesystem.free
            total:      system.filesystem.total
            available:  system.filesystem.available
            used_pct:   system.filesystem.used.pct
            used_bytes: system.filesystem.used.bytes

  - name: metricset-network

      fields:
        metricset.module: "system"
        metricset.name: "filesystem"

      buckets:
        terms:
          - host.name
          - system.network.name

      aggregate:
        terms:
            - tags
        stats:
            - system.network.in.bytes
            - system.network.in.dropped
            - system.network.in.errors
            - system.network.in.packats

            - system.network.out.bytes
            - system.network.out.dropped
            - system.network.out.errors
            - system.network.out.packats
```

```json
```


'''
arg_config = arg('--config', '-c', help="configuration file or '-' to read from stdin", default=None)

digest_command('query',
    arg_config,
    arg('--run-at', help="simulate running this program at given time"),
    arg('--starttime', help="start date"),
    arg('--endtime', help="end date"),
    arg('name', help="name of the digestion to run the query for"),
)
def cmd_query(config):
    from digest import Digester
    Digester.run_query(config)


digest_command('collect', arg_config,
    arg('--starttime', help="start date"),
    arg('--endtime', help="end date"),
    arg('--period', help="length of period (possible: 10s, 10m, 10h)"),
)
def cmd_collect(config): #, starttime, endtime, period):
    config = read_config(config)
    config['arguments'] = {}
    if starttime:
        config['arguments']['starttime'] = starttime

    if endtime:
        config['arguments']['endtime'] = endtime


import pyaml
@command('indices', arg_config)
def cmd_indices(config):
    from .connection import elasticsearch
    es = elasticsearch(config)
    pyaml.p([idx for idx in es.indices.get('_all').keys()])


@command('export',
    arg('--format', choices=('zip', 'dir'), default="dir"),
    arg('index_name', help="name of index to export"),
    )
def cmd_export(config):
    """Export indices into your working directory
    """
    from .connection import elasticsearch
    es = elasticsearch(config)
    from elasticsearch.helpers import scan

    index_name = config.get('export.index_name')
    result = es.indices.get(index_name)

    for idx_name, idx_data in result.items():

        if config.get('export.format') == 'dir':
            os.makedirs(idx_name)
            with open(join(idx_name, 'mappings.json'), 'w') as f:
                json.dump(result[idx_name]['mappings'], f)

            log.info("exporting %s to %s/", idx_name, idx_name)

            with open(join(idx_name, 'data.json'), 'w') as f:
                for data in scan(es, query={'query': {'match_all': {}}}, index=index_name):
                    json.dump(data, f)
                    f.write("\n")

        elif config.get('export.format') == 'zip':
            pass


@command('import', arg_config, arg('index_name'))
def cmd_import(config):
    index_name = config.get('index_name')

    with open(join(index_name, 'mappings.json'), 'r') as f:
        mappings = json.load(f)

    es.indices.create(index_name, body={'mappings': mappings})
    with open(join(index_name, 'data.json'), 'r') as f:
        bulk(es, [ json.loads(line) for line in f ], refresh=True)



from string import Formatter
import logging, json
log = logging.getLogger('elastico.formatter')

class ElasticoFormatter(Formatter):

    def format_field(self, value, format_spec):
        log.debug("format_field -- value=%r, format_spec=%r", value, format_spec)

        if format_spec.endswith('gb'):
            result = ('{:'+format_spec[:-2]+'f}').format(value/1000000000.0)
        elif format_spec.endswith('mb'):
            result = ('{:'+format_spec[:-2]+'f}').format(value/1000000.0)
        elif format_spec.endswith('json'):
            try:
                indent=int(format_spec[:-4])
            except:
                indent=None

            result = json.dumps(value, indent=indent)
        elif format_spec.endswith('indent'):
            try:
                first, rest = format_spec[:-6].split('.')
                first = " "*int(first)
                rest = " "*int(rest)
            except:
                first=''
                try:
                    rest = " "*int(format_spec[:-6])
                except:
                    rest=''

            result = indent(value, indent=rest)
        else:
            result = super(ElasticoFormatter, self).format_field(value, format_spec)

        log.debug("result=%r", result)

        return result

    def convert_field(self, value, conversion):
        if conversion == 'json':
            result = json.dumps(value, indent=2)
        else:
            parent = super(ElasticoFormatter, self)
            result = parent.convert_field(value, conversion)

        return result




"""This module provides ConfigFactory class.
"""

from .config import Config
import yaml, copy

from argdeco import ConfigDict
from os.path import exists, join

class ConfigFactory:
    """This class provides a factory producing a new Config object
    """

    def __init__(self, source=None):
        """Initialized usually from a config argument passed at commandline.

        :param source:
            If source is empty, :func:`create` will always produce an empty
            config object for the start (only `arguments` populated by given
            keywords.)

            If source is '-', data is read from stdin and parsed by YAML
            parser.  So you can pass YAML or JSON data.  This will be the
            initial content of the :class:`Config` object.

            If source is a filename and file exists, this file is read each
            time :func:`create` is run and parsed by YAML parser and a new
            a :class:`Config` object is produced from this data.
        """
        if isinstance(source, dict):
            self.config = source
        elif not source:
            self.config = {}
        elif source == '-':
            self.config = yaml.load(sys.stdin)
        elif exists(source):
            self.config = {}
            self.config_file = source

        self.arg_config = ConfigDict()

    def load_from_file(self, file_name):
        self.config_file = file_name

    def __setitem__(self, name, value):
        self.arg_config[name] = value

    # def update(self, *args, **kwargs):
    #     self.config.update(*args, **kwargs)

    def create(self, **kwargs):
        """Produce a new :class:`Config` object.

        :param **kwargs:
            Keyword arguments are used to populate config's `arguments`
            dictionary.
        """

        config = copy.deepcopy(self.config)
        config = Config(config)

        if hasattr(self, 'config_file'):
            config.include_file(self.config_file)
            config.set_filename(self.config_file)

        # #_action, _args, _kwargs = main.command.compile_args()
        #
        # action = main.command.get_action(name)
            #
        # for k,v in kwargs:
        #     main.command.get_config_name(action, v)

        #config['arguments'] = self.arg_config

        #config.update(self.arg_config)

        config['arguments'] = kwargs

        return config



import sys, yaml, os, pytz, pyaml, json
from os.path import exists, join, isdir
from subprocess import Popen, PIPE
from copy import deepcopy

if (sys.version_info > (3, 0)):
    PY3 = True
    string = str

else:
    PY3 = False
    string = basestring
    Exception = StandardError

import logging
log = logging.getLogger('elastico.util')

from datetime import datetime,date
from dateutil.parser import parse as dt_parse

def start_of_day(dt):
    return datetime.combine(to_dt(dt).date(), datetime.min.time())

def end_of_day(dt):
    return datetime.combine(to_dt(dt).date(), datetime.max.time())

def dt_isoformat(dt, sep='T', timespec='seconds'):
    if not isinstance(dt, (datetime, date)):
        dt = dt_parse(dt)

    try:
        result = dt.isoformat(sep, timespec)
        result = result.rsplit('+', 1)[0]
    except TypeError:
        result = dt.isoformat(sep)
        result = result.rsplit('+', 1)[0]

        if timespec == 'hours':
            result = result.split(':')[0]
        elif timespec == 'minutes':
            result = result.rsplit(':', 1)[0]
        elif timespec == 'seconds':
            if '.' in result:
                result = result.rsplit('.', 1)[0]
        else:
            raise Exception("timespec %s not supported", timespec)

    return result+"Z"

def to_dt(x):
    if not isinstance(x, datetime):
        x = dt_parse(x)
    if x.tzinfo is None:
        return pytz.UTC.localize(x)
    else:
        return x

def get_netrc_login_data(data, name):
    """
    raises LookupError, in case "name" not in "data"
    :returns:
    """
    # netrc configuration
    nrc = data.get(name, {})
    return get_netrc_login_data_from_value(nrc)

def get_netrc_login_data_from_value(nrc):
    if not nrc:
        raise LookupError("no netrc data present")

    if not isinstance(nrc, dict):
        filename = None
        machine  = nrc
    else:
        filename = nrc.get('file')
        machine  = nrc.get('machine')

    if machine is None:
        raise LookupError("no netrc data present")

    if nrc:
        import netrc
        (user, account, password) = netrc.netrc(filename).authenticators(machine)

    return (user, password)


def read_config_dir(path, config, name, recursive=False):
    '''read configuration files and extend config

    Read all yaml files from directory `path` (recursive) and extract all YAML
    documents (also multidocument YAML files) and append to configuration list
    named `name`.
    '''
    if name not in config:
        config[name] = []

    path = path.format(**config)

    if not exists(path): return

    if recursive:
        for root, dirs, files in os.walk(path):
            for fn in files:
                with open(join(root, fn), 'r') as f:
                    for _doc in yaml.load_all(f):
                        config[name].append(_doc)
    else:
        for fn in os.listdir(path):
            _fn = join(path, fn)
            if isdir(_fn): continue

            with open(_fn, 'r') as f:
                for _doc in yaml.load_all(f):
                    config[name].append(_doc)

def get_config_value(config, key, default=None):
    key_parts = key.split('.')
    try:
        result = format_value(config, config.get(key_parts[0], default))
    except Exception as e:
        log.debug("error in formatting %s", e)
        return default

    for k in key_parts[1:]:
        if k not in result:
            return default
        result = result[k]
    return result


def format_value(data, current=None):
    if current is None:
        current = data
    if isinstance(current, string):
        return current.format(**data)
    if isinstance(current, (list, tuple)):
        return [format_value(data, v) for v in current]
    if isinstance(current, dict):
        result = {}
        for k,v in current.items():
            result[k] = format_value(data, v)
        return result
    else:
        return current
    #
    # except Exception as e:
    #     log.debug("error formatting %s: %s", current, e)
    #     return default

def first_value(d):
    '''return the first value of dictionary d'''
    if PY3:
        return list(d.values())[0]
    else:
        return d.values()[0]

def write_output(config, data):
    output_format = config.get('output_format', 'yaml')
    if output_format == 'yaml':
        pyaml.p(data)
    elif output_format == 'json':
        print(json.dumps(data, indent=2))

def sendmail(host='localhost', port=0, use_ssl=False,
    username=None, password=None,
    sender=None, recipients=[], message=''):

    log.debug("sendmail")

    if use_ssl:
        from smtplib import SMTP_SSL as SMTP
    else:
        from smtplib import SMTP

    smtp = SMTP()
    smtp.connect(host=host, port=port)
    # if user and password are given, use them to smtp.login(user, pass)
    if username is not None:
        smtp.login(username, password)

    result = smtp.sendmail(sender, recipients, message)
    smtp.quit()
    return result


def run_command(kwargs, data=None):
    log = logging.getLogger('elastico.util.command')

    log.debug("run_command -- kwargs=%s", kwargs)

    if isinstance(kwargs, string):
        kwargs = {'args': kwargs, 'shell': True}
    elif isinstance(kwargs, (list, tuple)):
        kwargs = {'args': kwargs}
    else:
        kwargs = dict(kwargs)

    if isinstance(kwargs['args'], string):
        if 'shell' not in kwargs:
            kwargs['shell'] = True

    def _get_capture_value(name):
        if name in kwargs:
            return kwargs.pop(name)
        elif data is not None and name in data:
            return data[name]
        else:
            return False
        return

    capture_stdout = _get_capture_value('stdout')
    capture_stderr = _get_capture_value('stderr')

    if 'input' in kwargs:
        input = kwargs.pop('input')
        kwargs['stdin'] = PIPE
    else:
        input = None

    log.info("Popen -- kwargs=%s", kwargs)
    p = Popen(stdout=PIPE, stderr=PIPE, **kwargs)
    (stdout, stderr) = p.communicate(input)
    result = p.wait()

    if data is not None:
        _result = {}
        if capture_stdout:
            if stdout.count("\n".encode('utf-8')) == 1:
                stdout = stdout.strip()
            _result['stdout'] = stdout
        if capture_stderr:
            _result['stderr'] = stderr
        _result['exit_code'] = result

        data['result'] = _result

    return (result, stdout, stderr)


    log.debug("capture_stdout=%s, capture_stderr=%s", capture_stdout, capture_stderr)

    if rule is not None:
        if capture_stdout:
            if stdout.count("\n".encode('utf-8')) == 1:
                stdout = stdout.strip()
            rule['result.stdout'] = stdout
        if capture_stderr:
            rule['result.stderr'] = stderr
        rule['result.exit_code'] = result

    log.debug("rule: %s", rule)

    return (result, stdout, stderr)


def stripped(s, count=100):
    if len(s) > count:
        s = s[:count]+"..."
    return s

def get_alerts(_alerts, context):
    from .config import Config

    if isinstance(_alerts, dict):
        _tmp = []
        for k,v in _alerts.items():
            _value = Config.object({'type': context.format(k)})
            _value.update(deepcopy(v))
            _tmp.append(_value)
        _alerts = _tmp

    return _alerts

